VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
END
Attribute VB_Name = "RasterPixelCacher"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = False
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
Option Explicit

' this simple class is intended to wrap a raster and make
' spatial related pixel value access faster
' Version 1.0
' Modified 25/2/2008 by CJJ
' - added cached array variant array for writing to and reading from;
' - added buffered write functionality;
' - allow user to specify the desired size of the buffer.

Private Const DEFAULTWIDTH As Long = 150
Private Const DEFAULTHEIGHT As Long = 150

Private xMin As Double, xMax As Double, yMin As Double, yMax As Double
Private dblDX As Double, dblDY As Double
Private xMinCur As Double, xMaxCur As Double, yMinCur As Double, yMaxCur As Double
Private pR As IRaster
Private desiredWidth As Long, desiredHeight As Long
Private width As Long, height As Long
Private pPB As IPixelBlock
Private pPntCur As IPnt
Private writeArray As Variant
Private isDirty As Boolean

Public Property Let Raster(pRIn As IRaster)
    Dim pPnt As IPnt
    Dim pRProps As IRasterProps
        
    Set pR = pRIn
    Set pRProps = pRIn
    width = desiredWidth
    If pRProps.width < desiredWidth Then width = pRProps.width
    height = desiredHeight
    If pRProps.height < desiredHeight Then height = pRProps.height
    
    Set pPnt = New Pnt
    pPnt.SetCoords width, height
    Set pPB = pR.CreatePixelBlock(pPnt)
    Set pPnt = Nothing
    
    ' change the pixel type to something that VBA can handle (thanks VBA!)
    Dim i As Long
    For i = 0 To pPB.Planes - 1: pPB.PixelType(i) = PT_DOUBLE: Next i
    
    dblDX = pRProps.MeanCellSize.x
    dblDY = pRProps.MeanCellSize.y
    
    xMin = pRProps.Extent.xMin
    xMax = pRProps.Extent.xMax
    yMin = pRProps.Extent.yMin
    yMax = pRProps.Extent.yMax
    
    Set pRProps = Nothing
End Property

Public Property Get Raster() As IRaster
    Set Raster = pR
End Property

Public Property Get Value(plane As Long, x As Double, y As Double) As Double
    'Dim xTemp As Double, yTemp As Double
    Dim pPnt As IPnt
    Dim lngX As Long, lngY As Long
    
    If pR Is Nothing Or x < xMin Or x > xMax Or y < yMin Or y > yMax Then
        ' return error code
        Value = -9999
    ElseIf x < xMinCur Or x > xMaxCur Or y < yMinCur Or y > yMaxCur Then
'        ' read new values in and centre the array on the current location
'        lngX = Fix((x - xMin) / dblDX)
'        lngY = Fix((yMax - y) / dblDY)
'        lngX = lngX - Fix(width / 2)
'        lngY = lngY - Fix(height / 2)
'        ' check and correct to array bounds if required
'        ' calculate array bounds
'        xMinCur = xMin + lngX * dblDX
'        xMaxCur = xMinCur + width * dblDX
'        yMaxCur = yMax - lngY * dblDY
'        yMinCur = yMaxCur - height * dblDY
'        If xMinCur < xMin Then lngX = lngX + Fix((xMin - xMinCur) / dblDX)
'        If xMaxCur > xMax Then lngX = lngX - Fix((xMaxCur - xMax) / dblDX)
'        If yMinCur < yMin Then lngY = lngY - Fix((yMin - yMinCur) / dblDY)
'        If yMaxCur > yMax Then lngY = lngY + Fix((yMaxCur - yMax) / dblDY)
'        ' recalculate array bounds
'        xMinCur = xMin + lngX * dblDX
'        xMaxCur = xMinCur + width * dblDX
'        yMaxCur = yMax - lngY * dblDY
'        yMinCur = yMaxCur - height * dblDY
'
'        ' read array
'        Set pPnt = New Pnt
'        pPnt.SetCoords lngX, lngY
'        pR.Read pPnt, pPB
        
        WriteToRaster plane
        ReadFromRaster plane, x, y
        GoTo ReadValue
    Else
ReadValue:
        ' return value
        lngX = Fix((x - xMinCur) / dblDX)
        lngY = Fix((yMaxCur - y) / dblDY)
        Value = writeArray(lngX, lngY) 'pPB.GetVal(plane, lngX, lngY)
    End If

End Property

Public Sub SetValue(plane As Long, x As Double, y As Double, valueIn As Double)
    'Dim pPnt As IPnt
    Dim lngX As Long, lngY As Long
    
    If pR Is Nothing Or x < xMin Or x > xMax Or y < yMin Or y > yMax Then
        ' raise an error
        Err.Raise 1024, "RasterPixelCacher", "Outside raster extent"
    ElseIf x < xMinCur Or x > xMaxCur Or y < yMinCur Or y > yMaxCur Then
        ' save the already written current values to the raster if necessary
        WriteToRaster plane     ' the raster has been written to so save the changes
        ReadFromRaster plane, x, y
        GoTo WriteValue
    Else
WriteValue:
        ' write to the cache
        lngX = Fix((x - xMinCur) / dblDX)
        lngY = Fix((yMaxCur - y) / dblDY)
        writeArray(lngX, lngY) = valueIn
        isDirty = True
    End If
End Sub

Public Property Get dX() As Double
    If pR Is Nothing Then
        dX = -9999
    Else
        dX = dblDX
    End If
End Property

Public Property Get dY() As Double
    If pR Is Nothing Then
        dY = -9999
    Else
        dY = dblDY
    End If
End Property

Private Sub Class_Initialize()
    isDirty = False
    ' set up initial size of the buffer
    width = DEFAULTWIDTH
    height = DEFAULTHEIGHT
    desiredWidth = DEFAULTWIDTH
    desiredHeight = DEFAULTHEIGHT
End Sub

Private Sub ReadFromRaster(plane As Long, x As Double, y As Double)
    ' read new values in and centre the array on the current location
    Dim lngX As Long, lngY As Long
        
    lngX = Fix((x - xMin) / dblDX)
    lngY = Fix((yMax - y) / dblDY)
    lngX = lngX - Fix(width / 2)
    lngY = lngY - Fix(height / 2)
    ' check and correct to array bounds if required
    ' calculate array bounds
    xMinCur = xMin + lngX * dblDX
    xMaxCur = xMinCur + width * dblDX
    yMaxCur = yMax - lngY * dblDY
    yMinCur = yMaxCur - height * dblDY
    If xMinCur < xMin Then lngX = lngX + Fix((xMin - xMinCur) / dblDX)
    If xMaxCur > xMax Then lngX = lngX - Fix((xMaxCur - xMax) / dblDX)
    If yMinCur < yMin Then lngY = lngY - Fix((yMin - yMinCur) / dblDY)
    If yMaxCur > yMax Then lngY = lngY + Fix((yMaxCur - yMax) / dblDY)
    ' recalculate array bounds
    xMinCur = xMin + lngX * dblDX
    xMaxCur = xMinCur + width * dblDX
    yMaxCur = yMax - lngY * dblDY
    yMinCur = yMaxCur - height * dblDY
    
    ' read array
    Set pPntCur = New Pnt
    pPntCur.SetCoords lngX, lngY
    pR.Read pPntCur, pPB
    
    ' load the write cache
    writeArray = pPB.SafeArray(plane)
End Sub

Private Sub WriteToRaster(plane As Long)
    Dim pRE As IRasterEdit
    ' write the cached array to the raster
    
    ' exit the sub if there is nothing written to the write cache
    If Not isDirty Then Exit Sub
    
    pPB.SafeArray(plane) = writeArray
    Set pRE = pR
    If pRE.CanEdit Then
        ' write to same location that the pixels were read from
        pRE.Write pPntCur, pPB
    Else
        MsgBox "Raster cannot be written to", vbOKOnly, "RasterPixelCacher"
    End If
    
    ' reset the "Dirty" flag
    isDirty = False
End Sub

Public Sub Flush()
    WriteToRaster 0
End Sub

Private Sub Class_Terminate()
    WriteToRaster 0
End Sub

Public Sub SetDesiredBufferSize(NoCellsWide As Long, NoCellsHigh As Long)
    ' this method allows the user to specify a desired buffer size, which
    ' is then compared to the size of the raster, and a new buffer size is
    ' set, which will be used in the next raster read
    ' Carl Johnson (CJJ) 25/2/2008
    Dim pRProps As IRasterProps
    If NoCellsWide <> desiredWidth Then
        Set pRProps = pR
        desiredWidth = NoCellsWide
        width = desiredWidth
        If pRProps.width < desiredWidth Then width = pRProps.width
    End If
    If NoCellsHigh <> desiredHeight Then
        Set pRProps = pR
        desiredHeight = NoCellsHigh
        height = desiredHeight
        If pRProps.height < desiredHeight Then height = pRProps.height
    End If
    Set pRProps = Nothing
End Sub
